<?php
/*
 +--------------------------------------------------------------------+
 | Copyright CiviCRM LLC. All rights reserved.                        |
 |                                                                    |
 | This work is published under the GNU AGPLv3 license with some      |
 | permitted exceptions and without any warranty. For full license    |
 | and copyright information, see https://civicrm.org/licensing       |
 +--------------------------------------------------------------------+
 */

/**
 *
 * @package CRM
 * @copyright CiviCRM LLC https://civicrm.org/licensing
 */
class CRM_Contact_Form_Edit_TagsAndGroups {

  /**
   * Constant to determine which forms we are generating.
   *
   * Used by both profile and edit contact
   */
  const GROUP = 1, TAG = 2, ALL = 3;

  /**
   * Build form elements.
   *
   * @param CRM_Core_Form $form
   *   The form object that we are operating on.
   * @param int $contactId
   *   Contact id.
   * @param int $type
   *   What components are we interested in.
   * @param bool $visibility
   *   Visibility of the field.
   * @param null $isRequired
   * @param string $groupName
   *   If used for building group block.
   * @param string $tagName
   *   If used for building tag block.
   * @param string $fieldName
   *   This is used in batch profile(i.e to build multiple blocks).
   *
   * @param string $groupElementType
   *
   */
  public static function buildQuickForm(
    &$form,
    $contactId = 0,
    $type = self::ALL,
    $visibility = FALSE,
    $isRequired = NULL,
    $groupName = 'Group(s)',
    $tagName = 'Tag(s)',
    $fieldName = NULL,
    $groupElementType = 'checkbox'
  ) {
    if (!isset($form->_tagGroup)) {
      $form->_tagGroup = [];
    }

    // NYSS 5670
    if (!$contactId && !empty($form->_contactId)) {
      $contactId = $form->_contactId;
    }

    $type = (int) $type;
    if ($type & self::GROUP) {

      $fName = 'group';
      if ($fieldName) {
        $fName = $fieldName;
      }

      $groupID = isset($form->_grid) ? $form->_grid : NULL;
      if ($groupID && $visibility) {
        $ids = [$groupID => $groupID];
      }
      else {
        if ($visibility) {
          $group = CRM_Core_PseudoConstant::allGroup();
        }
        else {
          $group = CRM_Core_PseudoConstant::group();
        }
        $ids = $group;
      }

      if ($groupID || !empty($group)) {
        $groups = CRM_Contact_BAO_Group::getGroupsHierarchy($ids);

        // sudoman hack begins
        //
        // This hack allows us to filter Groups listings so users aren't
        // overwhelmed with choices.  The advantage of this over smart groups
        // is that normal groups use double opt-in.
        //
        // To set the variable, use the following method:
        //
        // $ drush vset --format=string groups_field_filters '
        // {
        //     "profile": {
        //         "468": {
        //             "groups": [41, 25],
        //             "titles": ["Yes, sign me up!", "Sign me up to the other group!"]
        //         }
        //     },
        // }'
        //
        // The "string" format must be used and not the "json" format.
        // Otherwise the json will be converted to a PHP array, which will not
        // work with this code.
        //
        //
        // The string contains json text, with the first level referring to
        // profiles, then the profile number, then the list of ids of groups to
        // display on that page and optional replacement titles for those
        // groups.


        // Find the list of groups that we want to let pass through a filter on
        // a given profile or page. If there is no filter for that page, then
        // all groups are displayed.

        $groups_field_filters = json_decode(variable_get('groups_field_filters', NULL), $assoc = TRUE);

        $form_class_name = get_class($form);
        switch ($form_class_name) {

          case 'CRM_Profile_Form_Edit':
            $profile_id = $form->get('gid'); // gid, not id
            $filter_config = $groups_field_filters['profile'][strval($profile_id)];
            break;

          default:
            foreach (['custom_pre_id', 'custom_post_id'] as $which_profile) {

              $profile_ids = $form->get('values')[$which_profile];

              // some pages can have multiple profiles
              if (!is_array($profile_ids)) {
                $all_profile_ids[] = $profile_ids;
              } else {
                $all_profile_ids = array_merge($all_profile_ids, $profile_ids);
              }
            }

            foreach ($all_profile_ids as $profile_id) {
              $possible_filter = $groups_field_filters['profile'][strval($profile_id)];

              if (isset($possible_filter)) {
                $filter_config = $possible_filter;
                break;
              }
            }
            break;
        }

        if (isset($filter_config)) {

          $filter_ids = $filter_config['groups'];
          $filter_titles = $filter_config['titles'];
        }
        // sudoman hack continues further below


        $attributes['skiplabel'] = TRUE;
        $elements = [];
        $groupsOptions = [];
        foreach ($groups as $id => $group) {

          // sudoman hack continues
          // filter groups if a filter is set
          if (isset($filter_ids) and !in_array($id, $filter_ids)) {
            continue;
          }
          // sudoman hack continues further below

          // make sure that this group has public visibility
          if ($visibility &&
            $group['visibility'] == 'User and User Admin Only'
          ) {
            continue;
          }

          if ($groupElementType == 'select') {
            $groupsOptions[$id] = $group['title'];
          }
          else {

            // sudoman hack continues
            if (isset($filter_titles)) {
              $custom_group_title = $filter_titles[array_search($id, $filter_ids)];
            }
            if (isset($custom_group_title)) {
              $group_title = $custom_group_title;
            } else {
              $group_title = $group['title'];
            }

            // only slight modification here
            $form->_tagGroup[$fName][$id]['description'] = $group['description'];
            $elements[] = &$form->addElement('advcheckbox', $id, NULL, $group_title, $attributes);

            // sudoman hack continues with calls to CRM_Contact_Form_Edit_TagsAndGroups::reInsertFilteredGroupMemberships
          }
        }

        if ($groupElementType == 'select' && !empty($groupsOptions)) {
          $form->add('select', $fName, $groupName, $groupsOptions, FALSE,
            ['id' => $fName, 'multiple' => 'multiple', 'class' => 'crm-select2 twenty']
          );
          $form->assign('groupCount', count($groupsOptions));
        }

        if ($groupElementType == 'checkbox' && !empty($elements)) {
          $form->addGroup($elements, $fName, $groupName, '&nbsp;<br />');
          $form->assign('groupCount', count($elements));
          if ($isRequired) {
            $form->addRule($fName, ts('%1 is a required field.', [1 => $groupName]), 'required');
          }
        }
        $form->assign('groupElementType', $groupElementType);
      }
    }

    if ($type & self::TAG) {
      $tags = CRM_Core_BAO_Tag::getColorTags('civicrm_contact');

      if (!empty($tags)) {
        $form->add('select2', 'tag', ts('Tag(s)'), $tags, FALSE, ['class' => 'huge', 'placeholder' => ts('- select -'), 'multiple' => TRUE]);
      }

      // build tag widget
      $parentNames = CRM_Core_BAO_Tag::getTagSet('civicrm_contact');
      CRM_Core_Form_Tag::buildQuickForm($form, $parentNames, 'civicrm_contact', $contactId, FALSE, TRUE);
    }
    $form->assign('tagGroup', $form->_tagGroup);
  }

  /**
   * Set defaults for relevant form elements.
   *
   * @param int $id
   *   The contact id.
   * @param array $defaults
   *   The defaults array to store the values in.
   * @param int $type
   *   What components are we interested in.
   * @param string $fieldName
   *   This is used in batch profile(i.e to build multiple blocks).
   *
   * @param string $groupElementType
   */
  public static function setDefaults($id, &$defaults, $type = self::ALL, $fieldName = NULL, $groupElementType = 'checkbox') {
    $type = (int ) $type;
    if ($type & self::GROUP) {
      $fName = 'group';
      if ($fieldName) {
        $fName = $fieldName;
      }

      $contactGroup = CRM_Contact_BAO_GroupContact::getContactGroup($id, 'Added', NULL, FALSE, TRUE);
      if ($contactGroup) {
        foreach ($contactGroup as $group) {
          if ($groupElementType == 'select') {
            $defaults[$fName][] = $group['group_id'];
          }
          else {
            $defaults[$fName . '[' . $group['group_id'] . ']'] = 1;
          }
        }
      }
    }

    if ($type & self::TAG) {
      $defaults['tag'] = implode(',', CRM_Core_BAO_EntityTag::getTag($id, 'civicrm_contact'));
    }
  }

  /**
   * Set default values for the form. Note that in edit/view mode
   * the default values are retrieved from the database
   *
   *
   * @param CRM_Core_Form $form
   * @param array $defaults
   */
  public static function setDefaultValues(&$form, &$defaults) {
    $contactEditOptions = $form->get('contactEditOptions');

    if ($form->_action & CRM_Core_Action::ADD) {
      if (array_key_exists('TagsAndGroups', $contactEditOptions)) {
        // set group and tag defaults if any
        if ($form->_gid) {
          $defaults['group'][$form->_gid] = 1;
        }
        if ($form->_tid) {
          $defaults['tag'][$form->_tid] = 1;
        }
      }
    }
    else {
      if (array_key_exists('TagsAndGroups', $contactEditOptions)) {
        // set the group and tag ids
        $groupElementType = 'checkbox';
        if (CRM_Utils_System::getClassName($form) == 'CRM_Contact_Form_Contact') {
          $groupElementType = 'select';
        }
        self::setDefaults($form->_contactId, $defaults, self::ALL, NULL, $groupElementType);
      }
    }
  }

  /** re-add groups selectively filtered out by sudoman's hack
   *
   * the sudoman hack continues from above and from calls to this function.
   *
   * re-add any filtered-out groups if the contact is already in those groups
   * so createProfileContact() doesn't remove group memberships.
   *
   * @param array   $profile_ids (ids of profiles in use on page)
   * @param int     $contactID (id of contact)
   * @param bool    $ignorePermission (passed to getContactGroup, often TRUE)
   * @param array   $params (form parameters)
   *
   * @return array  $params (modified parameters: active groups that were filtered are inserted)
   */
  public static function reInsertFilteredGroupMemberships($profile_ids,
      $contactID, $ignorePermission, $params) {

    $groups_field_filters = json_decode(variable_get('groups_field_filters', NULL), $assoc = TRUE);

    foreach ($profile_ids as $profile_id) {
      $filter = $groups_field_filters['profile'][strval($profile_id)];

      if (isset($filter)) {
        $filter_ids = $filter['groups'];
        break;
      }
    }

    if ($contactID !== NULL) {

      $contactGroupList = &CRM_Contact_BAO_GroupContact::getContactGroup($contactID, 'Added',
        NULL, FALSE, $ignorePermission
      );

      foreach ($contactGroupList as $value) {

        $groupId = $value['group_id'];
        if (!empty($filter_ids) && !in_array($groupId, $filter_ids)) {

          $params['group'][strval($groupId)] = 1;
        }
      }
    }

    return $params;
  }
  // sudoman hack ends

}
